/**
 * Copyright (C) 2010 Cloudfarming <info@cloudfarming.nl>
 *
 * Licensed under the Eclipse Public License - v 1.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *         http://www.eclipse.org/legal/epl-v10.html
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package nl.cloudfarming.client.geoviewer.jxmap.map;

import com.vividsolutions.jts.geom.Point;
import java.awt.EventQueue;
import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.util.Collections;
import java.util.HashMap;
import java.util.Map;
import java.util.Set;
import javax.swing.ActionMap;
import javax.swing.JPanel;
import javax.swing.text.DefaultEditorKit;
import nl.cloudfarming.client.geoviewer.BaseLayer;
import nl.cloudfarming.client.geoviewer.Category;
import nl.cloudfarming.client.geoviewer.EditLayer;
import nl.cloudfarming.client.geoviewer.Layer;
import nl.cloudfarming.client.geoviewer.events.LayerEventKey;
import nl.cloudfarming.client.geoviewer.LayerInfo;
import nl.cloudfarming.client.geoviewer.LayerObject;
import nl.cloudfarming.client.geoviewer.events.LayerObjectSelection;
import nl.cloudfarming.client.geoviewer.MapController;
import nl.cloudfarming.client.geoviewer.ObjectLayer;
import nl.cloudfarming.client.geoviewer.Preset;
import nl.cloudfarming.client.geoviewer.data.LayerDataManager;
import nl.cloudfarming.client.geoviewer.jxmap.layerlist.LayerListChildFactory;
import nl.cloudfarming.client.geoviewer.jxmap.layerlist.LayerListLayerNode;
import nl.cloudfarming.client.geoviewer.jxmap.layerlist.LayerListRootNode;
import nl.cloudfarming.eventbus.GuiEvent;
import nl.cloudfarming.eventbus.GuiEventBus;
import nl.cloudfarming.eventbus.GuiEventKey;
import nl.cloudfarming.eventbus.GuiEventListener;
import org.jdesktop.swingx.JXMapViewer;
import org.jdesktop.swingx.mapviewer.GeoPosition;
import org.openide.explorer.ExplorerManager;
import org.openide.explorer.ExplorerUtils;
import org.openide.nodes.Node;
import org.openide.util.Lookup;
import org.openide.util.lookup.Lookups;
import org.openide.util.lookup.ProxyLookup;

/**
 * Controller of the rootmap panel.
 * 
 *
 * @author Timon Veenstra
 */
public class RootMapPanelController implements MapController, Lookup.Provider, ExplorerManager.Provider {

    private final LayerDataManager layerDataManager;
    //TODO should root map panel get an explorer manager?
    private final RootMapPanel rootMapPanel;
    private final Map<Layer, LayerPanel> layerPanelMap = new HashMap<Layer, LayerPanel>();
    private boolean active = false;
    private final EditLayer editLayer = new EditLayer();
    private final ExplorerManager layerListEM = new ExplorerManager();
    private final Lookup lookup;

    public RootMapPanelController() {
        layerDataManager = LayerDataManager.getInstance();
        rootMapPanel = new RootMapPanel(layerDataManager.getExplorerDataProvider().getExplorerManager());
        //
        // initialize the layers
        //
        initLayers();


        layerListEM.setRootContext(new LayerListRootNode(new LayerListChildFactory(this)));
        layerListEM.getRootContext().setDisplayName("");

//        ExplorerChangeListener changeListener = new ExplorerChangeListener();
//        manager.addPropertyChangeListener(changeListener);

        ActionMap map = new ActionMap();
        map.put(DefaultEditorKit.copyAction, ExplorerUtils.actionCopy(layerListEM));
        map.put(DefaultEditorKit.cutAction, ExplorerUtils.actionCut(layerListEM));
        map.put(DefaultEditorKit.pasteAction, ExplorerUtils.actionPaste(layerListEM));
        map.put("delete", ExplorerUtils.actionDelete(layerListEM, true)); // or false

        //
        // proxy layerlist lookup aswel as lookup from the layer data manager to enable save cookie
        //
        this.lookup = new ProxyLookup(ExplorerUtils.createLookup(layerListEM, map), layerDataManager.getExplorerDataProvider().getLookup());

        layerListEM.addPropertyChangeListener(new LayerListSelectionChangeListener());
        //
        // register property change listener to handle change in node selection.
        //
        SelectionChangeListener changeListener = new SelectionChangeListener();
        GuiEventBus.addListener(changeListener);

        GuiEventBus.addListener(new NewLayerListener());
        GuiEventBus.addListener(new RemovedLayerListener());

        MapController.Registry.addMapController(this);

    }

    private void initLayers() {
        for (Layer layer : layerDataManager.getLayers()) {
            addLayer(layer);
        }
    }

    public JPanel getPanel() {
        return this.rootMapPanel;
    }

    private void repaintLayer(final Layer layer) {
        if (EventQueue.isDispatchThread()) {
            layerPanelMap.get(layer).repaint();
        } else {
            EventQueue.invokeLater(new Runnable() {

                @Override
                public void run() {
                    layerPanelMap.get(layer).repaint();
                }
            });
        }
    }

    private void removeLayer(final Layer layer) {
        LayerPanel layerPanel = layerPanelMap.get(layer);
        rootMapPanel.removeLayerPanel(layerPanel);
        layerPanelMap.remove(layer);
        refreshLayerList();
    }

    private void refreshLayerList() {
        if (this.layerListEM.getRootContext() instanceof LayerListRootNode) {
            ((LayerListRootNode) this.layerListEM.getRootContext()).refresh();
        }
    }

    private void addLayer(final Layer layer) {

        //
        // register listener for pallete change
        //
        layer.addPropertyChangeListener(BaseLayer.PROPERTY_PALETTE, new PropertyChangeListener() {

            @Override
            public void propertyChange(PropertyChangeEvent evt) {
                repaintLayer(layer);
            }
        });

        LayerPanel layerPanel;

        if (layer instanceof ObjectLayer) {
            layerPanel = new ObjectLayerPanel(this, rootMapPanel, (ObjectLayer) layer, getViewer());
            layer.addPropertyChangeListener(ObjectLayer.PROPERTY_OBJECTS, new PropertyChangeListener() {

                @Override
                public void propertyChange(PropertyChangeEvent evt) {
                    repaintLayer(layer);
                }
            });



        } else {
            layerPanel = new ImageLayerPanel(this, rootMapPanel, layer, getViewer());
        }



        layerPanelMap.put(layer, layerPanel);
        rootMapPanel.addLayerPanel(layerPanel);
        refreshLayerList();

    }

    @Override
    public Set<Layer> getLayers() {
        return Collections.unmodifiableSet(layerPanelMap.keySet());
    }

    @Override
    public void setLayerOrder(int[] order, Category category) {
        throw new UnsupportedOperationException("Not supported yet.");
    }

    @Override
    public LayerInfo getLayerInfo(Layer layer) {
        assert layerPanelMap.containsKey(layer);
        return layerPanelMap.get(layer).getInfo();
    }

    @Override
    public Preset getCurrentPreset() {
        throw new UnsupportedOperationException("Not supported yet.");
    }

    @Override
    public String getMapName() {
        return "map:" + System.identityHashCode(this);
    }

    JXMapViewer getViewer() {
        return rootMapPanel.getMapViewer();
    }

    @Override
    public void zoomToObject(LayerObject layerObject, Layer layer) {
        assert this.layerPanelMap.containsKey(layer);

        Point centroid = layerObject.getGeometry().getCentroid();
        GeoPosition geoPosition = new GeoPosition(centroid.getY(), centroid.getX());
        rootMapPanel.getMapViewer().setCenterPosition(geoPosition);
        rootMapPanel.getMapViewer().setZoom(3);
    }

    public boolean isActive() {
        return active;
    }

    public void setActive(boolean active) {
        this.active = active;
    }

    @Override
    public void setActiveLayer(Layer layer) {
        assert this.layerPanelMap.containsKey(layer);
        if (layer.isInterActive()) {
            rootMapPanel.setActiveLayer((ObjectLayerPanel) this.layerPanelMap.get(layer));
        } else {
            rootMapPanel.setActiveLayer(null);
        }
    }

    @Override
    public Lookup getLookup() {
        return this.lookup;
    }

    @Override
    public ExplorerManager getExplorerManager() {
        return this.layerListEM;
    }

    private class NewLayerListener implements GuiEventListener<Layer> {

        @Override
        public void onEvent(GuiEvent<Layer> event) {
            addLayer(event.getContent());
        }

        @Override
        public boolean listensTo(GuiEventKey guiEventKey) {
            return LayerEventKey.NEW_LAYER.equals(guiEventKey);
        }

        @Override
        public String getModuleName() {
            return "geoviewer-jxmap";
        }
    }

    private class RemovedLayerListener implements GuiEventListener<Layer> {

        @Override
        public void onEvent(GuiEvent<Layer> event) {
            removeLayer(event.getContent());
        }

        @Override
        public boolean listensTo(GuiEventKey guiEventKey) {
            return LayerEventKey.REMOVED_LAYER.equals(guiEventKey);
        }

        @Override
        public String getModuleName() {
            return "geoviewer-jxmap";
        }
    }

    private class SelectionChangeListener implements GuiEventListener<LayerObjectSelection> {

        @Override
        public void onEvent(GuiEvent<LayerObjectSelection> event) {
            Layer layer = event.getContent().getLayer();
            if (layer.isInterActive()) {
                layerPanelMap.get(layer).setSelectedObjects(event.getContent().getObjects());
            }
        }

        @Override
        public boolean listensTo(GuiEventKey guiEventKey) {
            return LayerEventKey.SELECTION_CHANGED.equals(guiEventKey);
        }

        @Override
        public String getModuleName() {
            return "geoviewer-jxmap";
        }
    }

    private class LayerListSelectionChangeListener implements PropertyChangeListener {

        @Override
        public void propertyChange(PropertyChangeEvent evt) {
            // handle change of node selection
            if (ExplorerManager.PROP_SELECTED_NODES.equals(evt.getPropertyName())) {
                Node[] nodes = (Node[]) evt.getNewValue();
                for (Node node : nodes) {
                    if (node instanceof LayerListLayerNode) {
                        setActiveLayer(((LayerListLayerNode) node).getLayer());
                    }
                }
            }
        }
    }
}
